function [forecasted_data,forecasted_fd,confidence_band]= forecast_continuation(past_data,current_data,options)
% FORECAST_CONTINUATION forecast the continuation of a curve given it
% beginning and past collected curves.
%  
% Arguments:
%     past_data: an pxn matrix where n is the number of past vectors and p is their length.
%     current_data: vector of length q (<p) of the beginning of the curve one whish to continue
%     options: a struct -
%     a) options for smoothing:
%       breakspast: a vector of the breaks (for knots) for smoothing the past data,
%                default-1:p
%       norder: order of the splines, default- 4 (cubic)%  
%       breaks_current: a vector of breaks (to locate knots) for  smoothing the current data
%                         default 1:q
%       lambda_smoothnesss: the penalty parameter controlling the smoothness of
%                the curves, default 1. 
%       number_pca : number of principal components that describe the past data
%                         default=3
%  
%    b) options for the forecasting:
%       lambda_ridge: the regularization constant for the forecasting, default=0.1
%       confidence_band: this is a flag for obtaining a confidence
%                            bands function. If confidence_band is 1 then
%                            the band be given in confidence_band and the band functions will be appended to the end
%                            of the forecasted_fd,default=0;
%       var_coef: a constant that control the width of the band. 
%               The band is given by 
%               [forecasted(t)-var_coef*std(forecasted(t)), forecasted(t)+var_coef*std(forecasted(t))]. 
%                Sould be obtained through cross validation. default 3
%
%       Last modified: 16 Novmber 2009

if(~exist('options') || ~isfield(options,'breaks_past')) options.breaks_past=1:size(past_data,1);end;
if(~isfield(options,'lambda_smoothnesss')) options.lambda_smoothnesss=1;end;
if(~isfield(options,'norder')) options.norder=4;end;
if(~isfield(options,'breaks_current')) options.breaks_current=1:size(current_data,1);end;
if(~isfield(options,'number_pca')) options.number_pca=1;end;
if(~isfield(options,'number_noise')) options.number_noise=10;end;

if(~isfield(options,'lambda_ridge')) options.lambda_ridge=0.1;end;
if(~isfield(options,'confidence_band')) options.confidence_band=0;end;
if(~isfield(options,'var_coef')) options.var_coef=3;end;



% creates curves from past data
breaks=options.breaks_past;
rng=[breaks(1),breaks(length(breaks))];
nbasis   = length(breaks) +options.norder - 2;
hbasis = create_bspline_basis(rng, nbasis, options.norder, breaks);
past_fdPar = fdPar(hbasis, 2,options.lambda_smoothnesss);
past_fd = smooth_basis(breaks, past_data, past_fdPar);

% computes the pca of past data. 
PCAfdPar = fdPar(hbasis, 2, options.lambda_smoothnesss);
pastPCAstr = pca_fd(past_fd, options.number_pca+options.number_noise, PCAfdPar);


% computes the current curve 
breaks_1=options.breaks_current;
if  (breaks_1 ~= breaks(1:length(breaks_1)))
    error('The knots of current data are different from those in the beginning of past data.');
end
range_1=[breaks_1(1),breaks_1(length(breaks_1))];
range_2=[breaks_1(length(breaks_1)),breaks(length(breaks))];
nbasis   = length(breaks_1) +options.norder - 2;
basis_1 = create_bspline_basis(range_1, nbasis, options.norder, breaks_1);
current_fdPar = fdPar(basis_1, 2, options.lambda_smoothnesss);
current_fd = smooth_basis(breaks_1, current_data, current_fdPar);


harm_fd = pastPCAstr.harmfd;
mean_fd = pastPCAstr.meanfd;


% restricts both the autocorrelation function and the mean to
% the beginning segment

harm_restricted1=restrict_bsplines_fd(harm_fd,range_1(1),range_1(2));
mean_restricted1=restrict_bsplines_fd(mean_fd,range_1(1),range_1(2));

harm_restricted2=restrict_bsplines_fd(harm_fd,range_2(1),range_2(2));
mean_restricted2=restrict_bsplines_fd(mean_fd,range_2(1),range_2(2));

basis_2 = getbasis(harm_restricted2);

% Defines the matrices A_1,and A_2, the loading matrices of the factors and
% L the factors variance matrix.

p=options.number_pca;
q=options.number_noise;

C_1=getcoef(harm_restricted1)';
A_1=C_1(1:p,:);
C_1=C_1(1:p+q,:);
C_2=getcoef(harm_restricted2)';
A_2=C_2(1:p,:);
C_2=C_2(1:p+q,:);

Lnoise=pastPCAstr.values;
Lnoise=diag(Lnoise(1:p+q));
L=diag(Lnoise(1:p));


% computes the coefficients for the forecated curve
sigma_square=options.lambda_ridge*sum(diag(L));
B1=(A_1*A_1'+sigma_square*pinv(L))\ A_1;
B=A_2'*B1;

coefs=getcoef(current_fd);
mean_restricted_coefs2=repmat(getcoef(mean_restricted2),1,size(coefs,2));
mean_restricted_coefs1=repmat(getcoef(mean_restricted1),1,size(coefs,2));


newcoefs=B*(coefs-mean_restricted_coefs1)+mean_restricted_coefs2;

% evals the forecasting at the knots
forecasted_fd=fd(newcoefs, basis_2, getnames(current_fd));
range_2  = getbasisrange(basis_2);
breaks_2 =  [range_2(1),...
    getbasispar(basis_2), range_2(2)];
forecasted_at_time_t=eval_fd(breaks_2,forecasted_fd , 0);
forecasted_data=forecasted_at_time_t(2:length(forecasted_at_time_t),:);

% computes the confidence band
confidence_band=[];
if(options.confidence_band==1)

Sigma=C_2'*(Lnoise-(C_1*C_1'+sigma_square*Lnoise^-1)\C_1*C_1'*Lnoise)* C_2; 
    %Sigma=A_2'*(L-B1*A_1'*L )*A_2;% the updated variance matrix.
    % the matrix C2{L- LC1(C1LC1'+sigmaI)^-1C1L) after manipulations
    I=eye(size(Sigma));

    
    Sigma_times_b_fd=fd(Sigma, basis_2, getnames(current_fd));
    I_times_b_fd=fd(I, basis_2, getnames(current_fd));
    
    Sigma_at_time_t=eval_fd(breaks_2, Sigma_times_b_fd, 0);
    I_at_time_t=eval_fd(breaks_2, I_times_b_fd, 0);
    
    % computes the band
    D= options.var_coef*sum(I_at_time_t.*Sigma_at_time_t,2).^(0.5);
    D=repmat(D,1,size(coefs,2));
    
    confidence_band=[forecasted_at_time_t-D,forecasted_at_time_t+D];
    % we shorten the band such that it will begin from the first unknown interval.
    confidence_band=confidence_band(2:size(confidence_band,1),:);

    % computes the curves of the bands
    band_fdPar = fdPar(basis_2, 2,options.lambda_smoothnesss);
    D_fd = smooth_basis(breaks_2, D, band_fdPar);
    upper_fd=plus(forecasted_fd,D_fd);
    lower_fd=minus(forecasted_fd,D_fd);
    forecasted_fd=horzcat(forecasted_fd,lower_fd);
    forecasted_fd=horzcat(forecasted_fd,upper_fd);


end


